/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/layer/utils/RasterContrastMenuWidget.cpp

  \brief This file defines a class for a RasterContrastMenuWidget.
*/

// TerraLib
#include "../../../../dataaccess/utils/Utils.h"
#include "../../../../maptools/RasterContrast.h"
#include "../../../../maptools/Utils.h"
#include "../../../../rp/Contrast.h"
#include "../../canvas/MapDisplay.h"
#include "../../utils/ScopedCursor.h"
#include "RasterContrastMenuWidget.h"

// Qt
#include <QActionGroup>

te::qt::widgets::RasterContrastMenuWidget::RasterContrastMenuWidget(te::qt::widgets::MapDisplay* mapDisplay, QWidget* parent, Qt::WindowFlags f): 
  QWidget(parent, f), 
  m_mapDisplay(mapDisplay)
{
  //create menu
  m_menu = new QMenu(tr("Auto Contrast"), this);

  //create actions
  QAction* actMinMax = new QAction(tr("Min Max"), this);
  QAction* act2Percent = new QAction(tr("2% Enhancement"), this);
  m_actClear = new QAction(tr("Clear"), this);
    
  //add actions to menu
  m_menu->addAction(actMinMax);
  m_menu->addAction(act2Percent);
  m_menu->addAction(m_actClear);

  //connections
  connect(actMinMax, SIGNAL(triggered()), this, SLOT(onContrastMinMaxClicked()));
  connect(act2Percent, SIGNAL(triggered()), this, SLOT(onContrast2PercentClicked()));
  connect(m_actClear, SIGNAL(triggered()), this, SLOT(onContrastClearClicked()));
}

te::qt::widgets::RasterContrastMenuWidget::~RasterContrastMenuWidget() = default;

void te::qt::widgets::RasterContrastMenuWidget::setLayer(te::map::AbstractLayerPtr layer)
{
  m_layer = layer;

  if (m_layer->getRasterContrast() == nullptr)
  {
    m_actClear->setEnabled(false);
  }
  else
  {
    m_actClear->setEnabled(true);
  }
}

QMenu* te::qt::widgets::RasterContrastMenuWidget::getMenu()
{
  return m_menu;
}

void te::qt::widgets::RasterContrastMenuWidget::onContrastMinMaxClicked()
{
  te::rst::Raster* raster = te::map::GetRaster(m_layer.get());

  if (raster != nullptr)
  {
    te::map::RasterContrast* contrast = new te::map::RasterContrast(te::map::RasterTransform::LINEAR_CONTRAST, raster->getNumberOfBands());
    
    te::rst::RasterSummary rasterSummary;
    if( ! te::rst::RasterSummaryManager::getInstance().get(
      raster, false, rasterSummary ) )
    {
      throw te::common::Exception(TE_TR("Raster summary Error"));
    }        

    for (int b = 0; b < static_cast<int>(raster->getNumberOfBands()); b++)
    {
      // get real min max values
      const std::complex<double>* cmin = rasterSummary.at(b).m_minVal;
      const std::complex<double>* cmax = rasterSummary.at(b).m_maxVal;
      double min = cmin->real();
      double max = cmax->real();

      double gain = 1.;
      double offset1 = 0.;
      double offset2 = 0.;

      if (te::rp::Contrast::getGainAndOffset(te::rp::Contrast::InputParameters::LinearContrastT, min, max, 0., 255., gain, offset1, offset2))
      {
        contrast->setValues(gain, offset1, offset2, min, max, b);
      }
    }

    m_layer->setRasterContrast(contrast);
  }

  delete raster;

  m_mapDisplay->updateLayer(m_layer, true);
}

void te::qt::widgets::RasterContrastMenuWidget::onContrast2PercentClicked()
{
  te::rst::Raster* raster = te::map::GetRaster(m_layer.get());

  if (raster != nullptr)
  {
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
    
    te::rst::RasterSummary rasterSummary;
    if( ! te::rst::RasterSummaryManager::getInstance().get(
      raster, false, rasterSummary ) )
    {
      throw te::common::Exception(TE_TR("Raster summary Error"));
    }            

    te::map::RasterContrast* contrast = new te::map::RasterContrast(te::map::RasterTransform::LINEAR_CONTRAST, raster->getNumberOfBands());

    for (int b = 0; b < static_cast<int>(raster->getNumberOfBands()); b++)
    {

      // get 2% histogram values
      std::map<double, unsigned> realHist( *rasterSummary.at(b).m_histogramR );

      double totalCount = 0.;

      for (std::map<double, unsigned>::iterator it = realHist.begin(); it != realHist.end(); ++it)
      {
        totalCount += it->second;
      }

      double minCount = totalCount * 0.02;
      double maxCount = totalCount * 0.98;

      double count = 0.;
      double min = 0.;
      double max = 0.;

      for (std::map<double, unsigned>::iterator it = realHist.begin(); it != realHist.end(); ++it)
      {
        count += it->second;

        if (count < minCount)
        {
          min = (*it).first;
        }

        if (maxCount <= count)
        {
          max = (*it).first;
          break;
        }
      }

      double gain = 1.;
      double offset1 = 0.;
      double offset2 = 0.;

      if (te::rp::Contrast::getGainAndOffset(te::rp::Contrast::InputParameters::LinearContrastT, min, max, 0., 255., gain, offset1, offset2))
      {
        contrast->setValues(gain, offset1, offset2, min, max, b);
      }
    }

    m_layer->setRasterContrast(contrast);
  }

  delete raster;

  m_mapDisplay->updateLayer(m_layer, true);
}

void te::qt::widgets::RasterContrastMenuWidget::onContrastClearClicked()
{
  m_layer->setRasterContrast(nullptr);

  m_mapDisplay->updateLayer(m_layer, true);
}
