/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/GeoMosaic.cpp
  \brief Create a mosaic from a set of geo-referenced rasters.
*/

#include "GeoMosaic.h"

#include "Macros.h"
#include "Functions.h"
#include "AlgorithmParametersSerializer.h"
#include "../raster/Interpolator.h"
#include "../raster/Enums.h"
#include "../raster/RasterFactory.h"
#include "../raster/Grid.h"
#include "../raster/Band.h"
#include "../raster/BandProperty.h"
#include "../raster/PositionIterator.h"
#include "../raster/Utils.h"
#include "../raster/Reprojection.h"
#include "../memory/CachedRaster.h"
#include "../geometry/Envelope.h"
#include "../geometry/GTFactory.h"
#include "../geometry/Polygon.h"
#include "../geometry/LinearRing.h"
#include "../geometry/MultiPolygon.h"
#include "../srs/Converter.h"
#include "../common/progress/TaskProgress.h"
#include "../common/MathUtils.h"

#include <boost/shared_array.hpp>
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>

#include <climits>
#include <cfloat>
#include <cmath>
#include <memory>
#include <cstdio>
#include <map>

namespace te
{
  namespace rp
  {

    GeoMosaic::InputParameters::InputParameters()
    {
      reset();
    }

    GeoMosaic::InputParameters::InputParameters( const InputParameters& other )
    {
      reset();
      operator=( other );
    }

    GeoMosaic::InputParameters::~InputParameters()
    {
      reset();
    }

    void GeoMosaic::InputParameters::reset() _NOEXCEPT_OP(false)
    {
      m_feederRasterPtr = nullptr;
      m_inputRastersBands.clear();
      m_interpMethod = te::rst::NearestNeighbor;
      m_noDataValue = 0.0;
      m_forceInputNoDataValue = false;
      m_blendMethod = te::rp::blender::NoBlendMethod;
      m_equalizationType = CenterEqType;
      m_useRasterCache = true;
      m_enableProgress = false;
      m_enableMultiThread = true;
      m_enableAutoValidAreaDetect = true;
      m_mosaicTargetMeans.clear();
      m_mosaicTargetStdDevs.clear();
      m_blenderRulePtr = nullptr;
      m_abortOnFirstError = true;
      m_skipInsideRasters = false;
      m_localEqMaxNeighborRasters = 4;
    }

    const GeoMosaic::InputParameters& GeoMosaic::InputParameters::operator=(
      const GeoMosaic::InputParameters& params )
    {
      reset();

      m_feederRasterPtr = params.m_feederRasterPtr;
      m_inputRastersBands = params.m_inputRastersBands;
      m_interpMethod = params.m_interpMethod;
      m_noDataValue = params.m_noDataValue;
      m_forceInputNoDataValue = params.m_forceInputNoDataValue;
      m_blendMethod = params.m_blendMethod;
      m_equalizationType = params.m_equalizationType;
      m_useRasterCache = params.m_useRasterCache;
      m_enableProgress = params.m_enableProgress;
      m_enableMultiThread = params.m_enableMultiThread;
      m_enableAutoValidAreaDetect = params.m_enableAutoValidAreaDetect;
      m_mosaicTargetMeans = params.m_mosaicTargetMeans;
      m_mosaicTargetStdDevs = params.m_mosaicTargetStdDevs;
      m_blenderRulePtr = params.m_blenderRulePtr;
      m_abortOnFirstError = params.m_abortOnFirstError;
      m_skipInsideRasters = params.m_skipInsideRasters;
      m_localEqMaxNeighborRasters = params.m_localEqMaxNeighborRasters;

      return *this;
    }

    te::common::AbstractParameters* GeoMosaic::InputParameters::clone() const
    {
      return new InputParameters( *this );
    }

    bool GeoMosaic::InputParameters::serialize( AlgorithmParametersSerializer& serializer ) const
    {
      TERP_TRUE_OR_RETURN_FALSE( AlgorithmInputParameters::serialize( serializer ),
        "Serializarion error" );

      serializer.setAlgorithmName( "GeoMosaic" );

      std::vector< std::string > rasterNames;
      m_feederRasterPtr->getRasterNames( rasterNames );
      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "inputRasters",
        rasterNames ), "Error adding parameter inputRasters" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameters( "inputRastersBands",
        m_inputRastersBands ), "Error adding parameter inputRastersBands" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "interpMethod",
        m_interpMethod ), "Error adding parameter interpMethod" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "noDataValue",
        m_noDataValue ), "Error adding parameter noDataValue" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "forceInputNoDataValue",
        m_forceInputNoDataValue ), "Error adding parameter forceInputNoDataValue" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "blendMethod",
        m_blendMethod ), "Error adding parameter blendMethod" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "equalizationType",
        m_equalizationType ), "Error adding parameter equalizationType" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "enableAutoValidAreaDetect",
        m_enableAutoValidAreaDetect ), "Error adding parameter enableAutoValidAreaDetect" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "mosaicTargetMeans",
        m_mosaicTargetMeans ), "Error adding parameter mosaicTargetMeans" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "mosaicTargetStdDevs",
        m_mosaicTargetStdDevs ), "Error adding parameter mosaicTargetStdDevs" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "abortOnFirstError",
        m_abortOnFirstError ), "Error adding parameter abortOnFirstError" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "skipInsideRasters",
        m_skipInsideRasters ), "Error adding parameter skipInsideRasters" );
      
      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "localEqMaxNeighborRasters",
        m_localEqMaxNeighborRasters ), "Error adding parameter localEqMaxNeighborRasters" );      

      return true;
    }
    
    void GeoMosaic::InputParameters::getEqTypesInfo( 
      std::vector< std::pair< EqType, std::string > >& eqTypesInfo  )
    {
      eqTypesInfo.clear();
      eqTypesInfo.push_back( std::pair< EqType, std::string >( NoEqType,
        "NoEqType" ) );
      eqTypesInfo.push_back( std::pair< EqType, std::string >( CenterEqType,
        "CenterEqType" ) );
      eqTypesInfo.push_back( std::pair< EqType, std::string >( TargetEqtype,
        "TargetEqtype" ) );
      eqTypesInfo.push_back( std::pair< EqType, std::string >( LocalEqType,
        "LocalEqType" ) );
    }

    GeoMosaic::OutputParameters::OutputParameters()
    {
      reset();
    }

    GeoMosaic::OutputParameters::OutputParameters( const OutputParameters& other )
    {
      reset();
      operator=( other );
    }

    GeoMosaic::OutputParameters::~OutputParameters()
    {
      reset();
    }

    void GeoMosaic::OutputParameters::reset() _NOEXCEPT_OP(false)
    {
      m_rType.clear();
      m_rInfo.clear();
      m_outputRasterPtr.reset();
    }

    const GeoMosaic::OutputParameters& GeoMosaic::OutputParameters::operator=(
      const GeoMosaic::OutputParameters& params )
    {
      reset();

      m_rType = params.m_rType;
      m_rInfo = params.m_rInfo;
      m_outputRasterPtr.reset();;

      return *this;
    }

    te::common::AbstractParameters* GeoMosaic::OutputParameters::clone() const
    {
      return new OutputParameters( *this );
    }

    GeoMosaic::GeoMosaic()
    {
      reset();
    }

    GeoMosaic::~GeoMosaic() = default;

    bool GeoMosaic::execute( AlgorithmOutputParameters& outputParams )
      _NOEXCEPT_OP(false)
    {
      if( ! m_isInitialized ) return false;
      
      GeoMosaic::OutputParameters* outParamsPtr = dynamic_cast<
        GeoMosaic::OutputParameters* >( &outputParams );
      TERP_INSTANCE_TRUE_OR_THROW( outParamsPtr, "Invalid paramters" );
      
      assert( m_inputParameters.m_feederRasterPtr->getObjsCount() > 1 );
      
      // First pass: getting global mosaic info

      double mosaicXResolution = 0.0;
      double mosaicYResolution = 0.0;
      double mosaicLLX = DBL_MAX; // world coords
      double mosaicLLY = DBL_MAX; // world coords
      double mosaicURX = -1.0 * DBL_MAX; // world coords
      double mosaicURY = -1.0 * DBL_MAX; // world coords
      int mosaicSRID = 0;     
      te::rst::BandProperty mosaicBaseBandProperties( 0, 0, "" );

      // a map ordering raster feeder indexes to the respective rasters bounding box (under the mosaic world coords).
      std::map< unsigned int, std::unique_ptr< te::gm::Polygon > > rastersBBoxesPolsMap;               
      
      // a map ordereding the distance to the mosaic center to each raster feeder index
      std::multimap< double, unsigned int > centerDist2RasterIdxMap;

      {
        std::unique_ptr< te::common::TaskProgress > progressPtr;
        if( m_inputParameters.m_enableProgress )
        {
          progressPtr.reset( new te::common::TaskProgress );
          
          progressPtr->setTotalSteps( static_cast<int>( 
            m_inputParameters.m_feederRasterPtr->getObjsCount()) );
          
          progressPtr->setMessage( "Mosaic: reading input rasters info" );
        }        
        
        // a map from input rasters SRIDS and their frequencies
        std::map< int, unsigned int > sridsFreq;
        
        // a map from input rasters SRIDS and their best resolutions
        std::map< int, std::pair< double, double > > srid2ResolutionsMap;
        
        {
          te::rst::Raster const* inputRasterPtr = nullptr;
          unsigned int inputRasterIdx = 0;          
          bool isRasterValid = false;

          m_inputParameters.m_feederRasterPtr->reset();
          while( ( inputRasterPtr = m_inputParameters.m_feederRasterPtr->getCurrentObj() ) )
          {
            inputRasterIdx = m_inputParameters.m_feederRasterPtr->getCurrentOffset();
            
            // checking if the current raster is valid 
            
            isRasterValid = true;
            
            if( ( inputRasterPtr->getAccessPolicy() & te::common::RAccess ) == 0 )
            {
              isRasterValid = false;
            }
            
            if( inputRasterPtr->getSRID() == 0 )
            {
              isRasterValid = false;
            }

            for( std::vector< unsigned int >::size_type inputRastersBandsIdx = 0 ;
              inputRastersBandsIdx <
              m_inputParameters.m_inputRastersBands[ inputRasterIdx ].size() ;
              ++inputRastersBandsIdx )
            {
              const unsigned int& currBand =
                m_inputParameters.m_inputRastersBands[ inputRasterIdx ][ inputRastersBandsIdx ];

              if( currBand >= inputRasterPtr->getNumberOfBands() )
              {
                isRasterValid = false;
                break;
              }
            }
            
            if( isRasterValid )
            {
              ++sridsFreq[ inputRasterPtr->getSRID() ];  
            
              rastersBBoxesPolsMap[ inputRasterIdx ] =
                std::unique_ptr< te::gm::Polygon >(           
                static_cast<te::gm::Polygon*>(te::gm::GetGeomFromEnvelope(
                inputRasterPtr->getGrid()->getExtent(), 
                inputRasterPtr->getGrid()->getSRID() ) ) );
                
              std::map< int, std::pair< double, double > >::iterator 
                srid2ResolutionsMapIt = srid2ResolutionsMap.find( 
                inputRasterPtr->getGrid()->getSRID() );
              if( 
                  srid2ResolutionsMapIt
                  ==
                  srid2ResolutionsMap.end()
                )
              {
                srid2ResolutionsMap[ inputRasterPtr->getGrid()->getSRID() ] = 
                  std::pair< double, double >(
                  inputRasterPtr->getGrid()->getResolutionX(),
                  inputRasterPtr->getGrid()->getResolutionY() );
              }
              else
              {
                if( 
                    ( 
                      srid2ResolutionsMapIt->second.first
                      <= 
                      inputRasterPtr->getGrid()->getResolutionX()
                    )
                    &&
                    ( 
                      srid2ResolutionsMapIt->second.second
                      <= 
                      inputRasterPtr->getGrid()->getResolutionY()
                    )
                  )
                {
                  srid2ResolutionsMapIt->second.first = 
                    inputRasterPtr->getGrid()->getResolutionX();
                  srid2ResolutionsMapIt->second.second =
                    inputRasterPtr->getGrid()->getResolutionY();
                }
              }
            }
            else
            {
              const std::string message = "Invalid input raster : " +
                std::to_string( inputRasterIdx );
              if( m_inputParameters.m_abortOnFirstError )
              {
                TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
              }
              else
              {
                TERP_INSTANCE_LOGWARN( message );
              }
            }

            // move to the next raster
            
            if( m_inputParameters.m_enableProgress )
            {
              progressPtr->pulse();
              if( ! progressPtr->isActive() ) return false;
            }                  

            m_inputParameters.m_feederRasterPtr->moveNext();
          }
        }
        
        // defining the mosaic SRID
        
        {
          std::map< int, unsigned int >::const_iterator sridsFreqIt = 
            sridsFreq.begin();
          const std::map< int, unsigned int >::const_iterator sridsFreqItEnd = 
            sridsFreq.end();
          unsigned highestSRIDFreq = 0;
          
          
          while( sridsFreqIt != sridsFreqItEnd )
          {
            if( highestSRIDFreq < sridsFreqIt->second )
            {
              highestSRIDFreq = sridsFreqIt->second;
              mosaicSRID = sridsFreqIt->first;
            }
            
            ++sridsFreqIt;
          }
        }
        
        // defining mosaic resolution
        
        mosaicXResolution = srid2ResolutionsMap[ mosaicSRID ].first;
        mosaicYResolution = srid2ResolutionsMap[ mosaicSRID ].second;
      
        // mosaic area
        
        {
          te::gm::Polygon* currPolPtr = 0;
            
          std::map< unsigned int, std::unique_ptr< te::gm::Polygon > >::const_iterator 
            rastersBBoxesPolsMapIt = rastersBBoxesPolsMap.begin();
          const std::map< unsigned int, std::unique_ptr< te::gm::Polygon > >::const_iterator 
            rastersBBoxesPolsMapItEnd = rastersBBoxesPolsMap.end();
          
          while( rastersBBoxesPolsMapIt != rastersBBoxesPolsMapItEnd )
          {
            currPolPtr = rastersBBoxesPolsMapIt->second.get();
            
            currPolPtr->transform( mosaicSRID );
            
            mosaicLLX = std::min( mosaicLLX, currPolPtr->getMBR()->getLowerLeftX() );
            mosaicLLY = std::min( mosaicLLY, currPolPtr->getMBR()->getLowerLeftY() );

            mosaicURX = std::max( mosaicURX, currPolPtr->getMBR()->getUpperRightX() );
            mosaicURY = std::max( mosaicURY, currPolPtr->getMBR()->getUpperRightY() );                  
            
            ++rastersBBoxesPolsMapIt;
          }
        }
        
        // updating center distances map
        
        {
          const double mosaicCenterX = ( mosaicURX + mosaicLLX ) / 2.0;
          const double mosaicCenterY = ( mosaicURY + mosaicLLY ) / 2.0;
          double currDist = 0;
          double currXDist = 0;
          double currYDist = 0;
          te::gm::Coord2D currCenter;
          te::gm::Polygon* currPolPtr = 0;
          
          std::map< unsigned int, std::unique_ptr< te::gm::Polygon > >::const_iterator 
            rastersBBoxesPolsMapIt = rastersBBoxesPolsMap.begin();
          const std::map< unsigned int, std::unique_ptr< te::gm::Polygon > >::const_iterator 
            rastersBBoxesPolsMapItEnd = rastersBBoxesPolsMap.end();          
          
          while( rastersBBoxesPolsMapIt != rastersBBoxesPolsMapItEnd )
          {
            currPolPtr = rastersBBoxesPolsMapIt->second.get();
            
            currCenter = currPolPtr->getMBR()->getCenter();
            currXDist = currCenter.getX() - mosaicCenterX;
            currYDist = currCenter.getY() - mosaicCenterY;
            currDist = std::sqrt( ( currXDist * currXDist ) + ( currYDist *
              currYDist ) );
            
            centerDist2RasterIdxMap.insert( std::pair< double, unsigned int >(
              currDist, rastersBBoxesPolsMapIt->first ) );
            
            ++rastersBBoxesPolsMapIt;
          }
        }
        
        // base band props
        // using the mosac center raster as reference
        
        {
          unsigned int centerInputRasterIdx = centerDist2RasterIdxMap.begin()->second;
          if( ! m_inputParameters.m_feederRasterPtr->moveTo( centerInputRasterIdx ) )
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Error opening input raster ");
          }
          
          te::rst::Raster const* centerInputRasterPtr = 
            m_inputParameters.m_feederRasterPtr->getCurrentObj();
          
          mosaicBaseBandProperties = *centerInputRasterPtr->getBand( 
            m_inputParameters.m_inputRastersBands[ centerInputRasterIdx ][ 0 ] )->getProperty();
          if( m_inputParameters.m_forceInputNoDataValue )
          {
            mosaicBaseBandProperties.m_noDataValue = m_inputParameters.m_noDataValue;
          }
        }
        
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( !rastersBBoxesPolsMap.empty(),
          "No input rasters to process" );        
      }

      // creating the output raster
      
      te::rst::Raster* outputRasterPtr = nullptr;

      {
        te::rst::Grid* outputGrid = new te::rst::Grid( mosaicXResolution,
          mosaicYResolution,  new te::gm::Envelope( mosaicLLX, mosaicLLY, mosaicURX,
          mosaicURY ), mosaicSRID );
        
        mosaicBaseBandProperties.m_blkh = 512;
        mosaicBaseBandProperties.m_blkw = 512;
        mosaicBaseBandProperties.m_nblocksx = (int)std::ceil( 
          ((double)outputGrid->getNumberOfColumns() )
          /
          ((double)mosaicBaseBandProperties.m_blkw) );
        mosaicBaseBandProperties.m_nblocksy = (int)std::ceil( 
          ((double)outputGrid->getNumberOfRows() )
          /
          ((double)mosaicBaseBandProperties.m_blkh) );        
        
        std::vector< te::rst::BandProperty* > bandsProperties;
        for( std::vector< unsigned int >::size_type bandIdx = 0 ;  bandIdx <
          m_inputParameters.m_inputRastersBands[ 0 ].size() ; ++bandIdx )
        {
          bandsProperties.push_back( new te::rst::BandProperty( mosaicBaseBandProperties ) );
          bandsProperties[ bandIdx ]->m_colorInterp = te::rst::GrayIdxCInt;
        }

        AlgorithmParametersSerializer paramsSerializer;
        TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
          "Parameter serialization error" );

        std::map <std::string, std::string> metadata;
        paramsSerializer.getMetaData( metadata );

        try
        {
          outParamsPtr->m_outputRasterPtr.reset(
            te::rst::RasterFactory::make(
              outParamsPtr->m_rType,
              outputGrid,
              bandsProperties,
              outParamsPtr->m_rInfo,
              metadata,
              nullptr,
              nullptr ) );
        }
        catch(...)
        {
          outParamsPtr->m_outputRasterPtr.reset();
        }
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( outParamsPtr->m_outputRasterPtr.get(),
          "Output raster creation error" );
          
        outputRasterPtr = outParamsPtr->m_outputRasterPtr.get();
      }

      std::unique_ptr< te::mem::CachedRaster > outputCachedRasterInstancePtr;
      
      if( m_inputParameters.m_useRasterCache )
      {
        outputCachedRasterInstancePtr.reset( new te::mem::CachedRaster(
          *(outParamsPtr->m_outputRasterPtr.get()), 50, 0 ) );   
          
        outputRasterPtr = outputCachedRasterInstancePtr.get();
      }

      // fill output with no data values

      {
        const unsigned int nBands = static_cast<unsigned int>(outputRasterPtr->getNumberOfBands());
        unsigned int bandIdx = 0;

        for( bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
        {
          te::rst::Band& outBand = ( *( outputRasterPtr->getBand( bandIdx ) ) );
          
          const std::complex<double > bandNoDataValue( 
            outBand.getProperty()->m_noDataValue,
            outBand.getProperty()->m_noDataValue );
          
          te::rp::FillBand( *outputRasterPtr, bandIdx, bandNoDataValue );
        }
      }
      
      // iterating over the all rasters
      
      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr.reset( new te::common::TaskProgress );
        
        progressPtr->setTotalSteps( static_cast<int>( 
          centerDist2RasterIdxMap.size() ) );
        
        progressPtr->setMessage( "Mosaic: processing input rasters" );
      }              
      
      // the union of all blended rasters valid areas under mosaic SRID
      std::unique_ptr< te::gm::MultiPolygon > mosaicValidAreasUnionPtr;
      
      // rasters (that were blended into the mosaic) bounding boxes centers 
      // (mosaic SRID) to their updated statistics (mean, stddev) for 
      // each band
      std::vector< std::tuple< te::gm::Coord2D, std::vector< double >, 
        std::vector< double > > >  bBoxesCenters2Stats;
      
      std::vector< unsigned int > outputRasterBands;
      std::vector< double > dummyRasterOffsets;
      std::vector< double > dummyRasterScales;
      for( unsigned int outputRasterBandsIdx = 0 ; outputRasterBandsIdx <
        outputRasterPtr->getNumberOfBands() ; ++outputRasterBandsIdx )
      {
        outputRasterBands.push_back( outputRasterBandsIdx );
        dummyRasterOffsets.push_back( 0.0 );
        dummyRasterScales.push_back( 1.0 );
      }      
      
      std::vector< double > mosaicTargetMeans;
      std::vector< double > mosaicTargetStdDevs;      

      std::multimap< double, unsigned int >::const_iterator
        centerDist2RasterIdxMapIt = centerDist2RasterIdxMap.begin();
      const std::multimap< double, unsigned int >::const_iterator
        centerDist2RasterIdxMapItEnd = centerDist2RasterIdxMap.end();
      
      while( centerDist2RasterIdxMapIt != centerDist2RasterIdxMapItEnd )
      {
        if( m_inputParameters.m_enableProgress )
        {
          progressPtr->pulse();
          if( ! progressPtr->isActive() ) return false;
        }        
        
        const unsigned int inputRasterIdx = centerDist2RasterIdxMapIt->second;
        
        // is this an inside raster ?
        
        if( 
            m_inputParameters.m_skipInsideRasters
            &&
            ( mosaicValidAreasUnionPtr.get() != nullptr )
          )
        {
          try
          {
            if( 
                mosaicValidAreasUnionPtr->covers( rastersBBoxesPolsMap[ inputRasterIdx ].get() ) 
                ||
                mosaicValidAreasUnionPtr->contains( rastersBBoxesPolsMap[ inputRasterIdx ].get() )
              )
            {
              const std::string message = "Skipping interior raster " +
                std::to_string( inputRasterIdx );                 
              TERP_INSTANCE_LOGWARN( message );
              ++centerDist2RasterIdxMapIt;
              continue;
            }
          }
          catch(...)
          {
            const std::string message = "Unable to find out raster bounding boxes relation " +
              std::to_string( inputRasterIdx );            
            if( m_inputParameters.m_abortOnFirstError )
            {
              TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
            }
            else
            {
              TERP_INSTANCE_LOGWARN( message );
              ++centerDist2RasterIdxMapIt;
              continue;
            }
          }
        }
        
        if( ! m_inputParameters.m_feederRasterPtr->moveTo( inputRasterIdx ) )
        {
          const std::string message = "Error opening input raster " +
            std::to_string( inputRasterIdx );
          if( m_inputParameters.m_abortOnFirstError )
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
          }
          else
          {
            TERP_INSTANCE_LOGWARN( message );
            ++centerDist2RasterIdxMapIt;
            continue;
          }
        }
        
        te::rst::Raster const* inputRasterPtr = m_inputParameters.m_feederRasterPtr->getCurrentObj();

        // reprojection 
        
        std::unique_ptr< te::rst::Raster > reprojectedInputRasterPtr;
        
        if( outputRasterPtr->getSRID() != inputRasterPtr->getSRID() )
        {
          std::vector< unsigned int > inputRasterBands;
          for( unsigned int bandIdx = 0 ; bandIdx < inputRasterPtr->getNumberOfBands() ;
            ++bandIdx )
          {
            inputRasterBands.push_back( bandIdx );
          }
          
          const bool repResult =
            te::rst::Reproject(
              outputRasterPtr->getSRID(),
              *inputRasterPtr,
              inputRasterBands,
              m_inputParameters.m_interpMethod,
              1, // interpWindowRadius
              0, // firstRow
              0, // firstColumn
              inputRasterPtr->getNumberOfRows(),
              inputRasterPtr->getNumberOfColumns(),
              inputRasterPtr->getNumberOfRows(),
              inputRasterPtr->getNumberOfColumns(),
              0, // maxThreads
              std::map< std::string, std::string >(),
              "EXPANSIBLE",
              reprojectedInputRasterPtr
            );
          
          if( !repResult )
          {
            const std::string message = "Error reprojecting input raster " +
              std::to_string( inputRasterIdx );
            if( m_inputParameters.m_abortOnFirstError )
            {
              TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
            }
            else
            {
              TERP_INSTANCE_LOGWARN( message );
              ++centerDist2RasterIdxMapIt;
              continue;
            }
          }
           
          inputRasterPtr = reprojectedInputRasterPtr.get();
        }
        
        // Caching issues 
        
        std::unique_ptr< te::mem::CachedRaster > cachedInputRasterPtr;
        
        if(
            m_inputParameters.m_useRasterCache
           &&
           ( reprojectedInputRasterPtr.get() == nullptr ) // Reprojected rasters already have internal cache
          )
        {
          cachedInputRasterPtr.reset( new te::mem::CachedRaster(
            *inputRasterPtr, 20, 0 ) );
          inputRasterPtr = cachedInputRasterPtr.get();
        }         
        
        // The transformation mapping outputRaster pixels 
        // ( te::gm::GTParameters::TiePoint::first ) to input pixels 
        // ( te::gm::GTParameters::TiePoint::second ) 
        // (Note: all coords are indexed by lines/columns).
        
        std::unique_ptr< te::gm::GeometricTransformation > transPtr;
        
        {
          te::gm::GTParameters transParams;
          double auxX;
          double auxY;
          te::gm::GTParameters::TiePoint auxTP;
          
          auxTP.first.x = 0;
          auxTP.first.y = 0;
          outputRasterPtr->getGrid()->gridToGeo( auxTP.first.x, auxTP.first.y, auxX, auxY );
          inputRasterPtr->getGrid()->geoToGrid( auxX, auxY, auxTP.second.x, auxTP.second.y );
          transParams.m_tiePoints.push_back( auxTP );
          
          auxTP.first.x = 0;
          auxTP.first.y = static_cast<double>( outputRasterPtr->getNumberOfRows() - 1 );
          outputRasterPtr->getGrid()->gridToGeo( auxTP.first.x, auxTP.first.y, auxX, auxY );
          inputRasterPtr->getGrid()->geoToGrid( auxX, auxY, auxTP.second.x, auxTP.second.y );          
          transParams.m_tiePoints.push_back( auxTP );

          auxTP.first.x = static_cast<double>( outputRasterPtr->getNumberOfColumns() - 1);
          auxTP.first.y = 0;
          outputRasterPtr->getGrid()->gridToGeo( auxTP.first.x, auxTP.first.y, auxX, auxY );
          inputRasterPtr->getGrid()->geoToGrid( auxX, auxY, auxTP.second.x, auxTP.second.y );
          transParams.m_tiePoints.push_back( auxTP );
          
          auxTP.first.x = static_cast<double>( outputRasterPtr->getNumberOfColumns() - 1 );
          auxTP.first.y = static_cast<double>( outputRasterPtr->getNumberOfRows() - 1 );
          outputRasterPtr->getGrid()->gridToGeo( auxTP.first.x, auxTP.first.y, auxX, auxY );
          inputRasterPtr->getGrid()->geoToGrid( auxX, auxY, auxTP.second.x, auxTP.second.y );                    
          transParams.m_tiePoints.push_back( auxTP );
          
          transPtr.reset( te::gm::GTFactory::make( "Affine" ) );
          TERP_INSTANCE_TRUE_OR_RETURN_FALSE( transPtr.get(), "Could not instantiate a geometric transformation" );
          
          TERP_INSTANCE_TRUE_OR_RETURN_FALSE( transPtr->initialize( transParams ),
            "Could not initialize a geometric transformation" );
        }

        // Locating the raster valid data area if required
        
        std::unique_ptr< te::gm::MultiPolygon >
          currRasterValidDataDelimiterMultiPolPtr;
        
        {
          std::unique_ptr< te::gm::Polygon > 
            currRasterValidDataDelimiterPolygonPtr;          
          
          if( m_inputParameters.m_enableAutoValidAreaDetect )
          {
            std::vector< double > bandNoDataValues;          
            if( m_inputParameters.m_forceInputNoDataValue )
            {
              bandNoDataValues.resize( m_inputParameters.m_inputRastersBands[ 
                inputRasterIdx ].size(), m_inputParameters.m_noDataValue );
            }
            
            std::unique_ptr< te::gm::Polygon > outPolygonPtr;
            
            if( te::rp::GetExternalValidDataPolygon( *inputRasterPtr, 
              m_inputParameters.m_inputRastersBands[ inputRasterIdx ],
              bandNoDataValues, outPolygonPtr ) )
            {
              TERP_DEBUG_TRUE_OR_THROW( outPolygonPtr->isValid(), "Invalid valid area polygon" );
              
              currRasterValidDataDelimiterPolygonPtr.reset( outPolygonPtr.release() );
            }
            else
            {
              const std::string message = "Error calculating input raster valid data area " +
                std::to_string( inputRasterIdx );
              if( m_inputParameters.m_abortOnFirstError )
              {
                TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
              }
              else
              {
                TERP_INSTANCE_LOGWARN( message );
                ++centerDist2RasterIdxMapIt;
                continue;
              }
            }
          }
          else
          {
            currRasterValidDataDelimiterPolygonPtr.reset(
              static_cast<te::gm::Polygon*>(te::gm::GetGeomFromEnvelope(
              inputRasterPtr->getGrid()->getExtent(), 
              inputRasterPtr->getGrid()->getSRID() ) ) );
          }
          
          currRasterValidDataDelimiterMultiPolPtr.reset( new te::gm::MultiPolygon( 
            0, te::gm::MultiPolygonType, 
            currRasterValidDataDelimiterPolygonPtr->getSRID(), nullptr ) );
          currRasterValidDataDelimiterMultiPolPtr->add( 
            currRasterValidDataDelimiterPolygonPtr.release() );
          currRasterValidDataDelimiterMultiPolPtr->computeMBR( true );
          TERP_DEBUG_TRUE_OR_THROW( currRasterValidDataDelimiterMultiPolPtr->isValid(),
            "Invalid valid area multi-polygon");
        }
        
        // equalization stuff

        std::vector< double > currentRasterBandsOffsets;
        std::vector< double > currentRasterBandsScales;

        if( m_inputParameters.m_equalizationType == InputParameters::NoEqType )
        {
          currentRasterBandsOffsets = dummyRasterOffsets;
          currentRasterBandsScales = dummyRasterScales;
        }
        else
        {
          std::vector< double > currentRasterMeans( 
            m_inputParameters.m_inputRastersBands[ inputRasterIdx ].size(), 0 );
          std::vector< double > currentRasterStdDevs( 
            m_inputParameters.m_inputRastersBands[ inputRasterIdx ].size(), 0 );

          for( unsigned int inputRastersBandsIdx = 0 ; inputRastersBandsIdx <
            m_inputParameters.m_inputRastersBands[ inputRasterIdx ].size() ;
            ++inputRastersBandsIdx )
          {
            const unsigned int inputBandIdx = m_inputParameters.m_inputRastersBands[ inputRasterIdx ][
              inputRastersBandsIdx ];
              
            if( !
                (
                  te::rp::GetMeanValue( 
                    *inputRasterPtr,
                    inputBandIdx, 
                    0, // auto number of threads
                    m_inputParameters.m_forceInputNoDataValue, 
                    m_inputParameters.m_noDataValue, 
                    currentRasterMeans[ inputRastersBandsIdx ] 
                  )
                  &&
                  te::rp::GetStdDevValue( 
                    *inputRasterPtr,
                    inputBandIdx, 
                    0, // auto number of threads
                    m_inputParameters.m_forceInputNoDataValue, 
                    m_inputParameters.m_noDataValue, 
                    &(currentRasterMeans[ inputRastersBandsIdx ]),
                    currentRasterStdDevs[ inputRastersBandsIdx ] 
                  )
                )
              )
            {
              const std::string message = "Error calculating statistics from input raster " +
                std::to_string( inputRasterIdx );
              if( m_inputParameters.m_abortOnFirstError )
              {
                TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
              }
              else
              {
                TERP_INSTANCE_LOGWARN( message );
                ++centerDist2RasterIdxMapIt;
                continue;
              }
            }
          }
          
          switch( m_inputParameters.m_equalizationType )
          {
            case InputParameters::CenterEqType :
            {
              if( mosaicTargetMeans.empty() )
              {
                mosaicTargetMeans = currentRasterMeans;
                mosaicTargetStdDevs = currentRasterStdDevs;
              }
              
              break;
            }
            case InputParameters::TargetEqtype :
            {
              if( mosaicTargetMeans.empty() )
              {
                mosaicTargetMeans = m_inputParameters.m_mosaicTargetMeans;
                mosaicTargetStdDevs = m_inputParameters.m_mosaicTargetStdDevs;
              }
              
              break;
            }
            case InputParameters::LocalEqType :
            {
              if( bBoxesCenters2Stats.empty() )
              {
                mosaicTargetMeans = currentRasterMeans;
                mosaicTargetStdDevs = currentRasterStdDevs;
              }
              else
              {
                const te::gm::Coord2D& currRasterCenter =
                  inputRasterPtr->getGrid()->getExtent()->getCenter();
                const std::size_t bBoxesCenters2StatsSize = 
                  bBoxesCenters2Stats.size();
                std::multimap< double, std::size_t > dist2IdxMap;
                double dist = 0;
                double diffX = 0;
                double diffY = 0;

                for( std::size_t bBoxesCenters2StatsIdx = 0 ; bBoxesCenters2StatsIdx 
                  < bBoxesCenters2StatsSize ; ++bBoxesCenters2StatsIdx )
                {
                  diffX = std::get<0 >( bBoxesCenters2Stats[ 
                    bBoxesCenters2StatsIdx ] ).x - currRasterCenter.x;
                  diffY = std::get<0 >( bBoxesCenters2Stats[ 
                    bBoxesCenters2StatsIdx ] ).y - currRasterCenter.y;
                  dist = std::sqrt( ( diffX * diffX ) + ( diffY * diffY ) );
                  
                  dist2IdxMap.insert( std::pair< double, std::size_t >(
                    dist, bBoxesCenters2StatsIdx ) );
                }
                
                std::fill( mosaicTargetMeans.begin(), mosaicTargetMeans.end(), 0 );
                std::fill( mosaicTargetStdDevs.begin(), mosaicTargetStdDevs.end(), 0 );
                
                unsigned int nRastersNmb = 0;
                std::multimap< double, std::size_t >::const_iterator 
                  dist2IdxMapIt = dist2IdxMap.begin();
                const std::multimap< double, std::size_t >::const_iterator 
                  dist2IdxMapItEnd = dist2IdxMap.end();                  
                
                while( 
                        ( dist2IdxMapIt != dist2IdxMapItEnd )
                        &&
                        ( nRastersNmb < m_inputParameters.m_localEqMaxNeighborRasters )
                     )
                {
                  const std::size_t bBoxesCenters2StatsIdx = dist2IdxMapIt->second;
                  const std::vector< double >& nRasterMeans = 
                    std::get< 1 >( bBoxesCenters2Stats[ bBoxesCenters2StatsIdx ] ); 
                  const std::vector< double >& nRasterStdDevs = 
                    std::get< 2 >( bBoxesCenters2Stats[ bBoxesCenters2StatsIdx ] );
                  
                  for( unsigned int outputRasterBandsIdx = 0 ; outputRasterBandsIdx <
                    outputRasterPtr->getNumberOfBands() ; ++outputRasterBandsIdx )
                  {
                    mosaicTargetMeans[ outputRasterBandsIdx ] +=
                      nRasterMeans[ outputRasterBandsIdx ];
                    mosaicTargetStdDevs[ outputRasterBandsIdx ] +=
                      nRasterStdDevs[ outputRasterBandsIdx ];                      
                  }                        
                  
                  ++nRastersNmb;
                  ++dist2IdxMapIt;
                }
                
                for( unsigned int outputRasterBandsIdx = 0 ; outputRasterBandsIdx <
                  outputRasterPtr->getNumberOfBands() ; ++outputRasterBandsIdx )
                {
                  mosaicTargetMeans[ outputRasterBandsIdx ] /=
                    (double)nRastersNmb;
                  mosaicTargetStdDevs[ outputRasterBandsIdx ] /=
                    (double)nRastersNmb;
                }                        
              }
              
              break;
            }            
            default :
            {
              TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Invalid equalization type");
              break;
            }
          }

          for( unsigned int inputRastersBandsIdx = 0 ; inputRastersBandsIdx <
            m_inputParameters.m_inputRastersBands[ inputRasterIdx ].size() ;
            ++inputRastersBandsIdx )
          {
            if( currentRasterStdDevs[ inputRastersBandsIdx ] == 0 )
            {
              currentRasterBandsScales.push_back( 1.0 );
              currentRasterBandsOffsets.push_back( 0.0 );
            }
            else
            {
              currentRasterBandsScales.push_back( 
                ( 
                  std::sqrt( mosaicTargetStdDevs[ inputRastersBandsIdx ] )
                   /
                  std::sqrt( currentRasterStdDevs[ inputRastersBandsIdx ] ) 
                )
              );
              currentRasterBandsOffsets.push_back(
                ( 
                  mosaicTargetMeans[ inputRastersBandsIdx ]
                   - 
                  ( 
                    currentRasterBandsScales[ inputRastersBandsIdx ] 
                    * 
                    currentRasterMeans[ inputRastersBandsIdx ] 
                  ) 
                )
              );
            }
            
            currentRasterMeans[ inputRastersBandsIdx ] = 
              ( 
                currentRasterMeans[ inputRastersBandsIdx ]
                *
                currentRasterBandsScales[ inputRastersBandsIdx ]
              )
              +
              currentRasterBandsOffsets[ inputRastersBandsIdx ]
              ;
              
            currentRasterStdDevs[ inputRastersBandsIdx ] *= 
              ( 
                currentRasterBandsScales[ inputRastersBandsIdx ]
                *
                currentRasterBandsScales[ inputRastersBandsIdx ]
              );
          }
          
          std::tuple< te::gm::Coord2D, std::vector< double >, 
            std::vector< double > > auxTuple;
          std::get<0 >( auxTuple ) = inputRasterPtr->getGrid()->getExtent()->getCenter();
          std::get<1 >( auxTuple ) = currentRasterMeans;
          std::get<2 >( auxTuple ) = currentRasterStdDevs;              
          bBoxesCenters2Stats.push_back( auxTuple );          
        }
        
         // blending
         
        {
          te::rp::Blender blenderInstance;

          if( 
              !
              ( 
                (
                  blenderInstance.initialize(
                  *outputRasterPtr,
                  outputRasterBands,
                  *inputRasterPtr,
                  m_inputParameters.m_inputRastersBands[ inputRasterIdx ],
                  m_inputParameters.m_blendMethod,
                  te::rst::NearestNeighbor,
                  m_inputParameters.m_interpMethod,
                  m_inputParameters.m_noDataValue,
                  false,
                  m_inputParameters.m_forceInputNoDataValue,
                  dummyRasterOffsets,
                  dummyRasterScales,
                  currentRasterBandsOffsets,
                  currentRasterBandsScales,
                  mosaicValidAreasUnionPtr.get(),
                  currRasterValidDataDelimiterMultiPolPtr.get(),
                  *transPtr,
                  0,
                  false,
                  m_inputParameters.m_blenderRulePtr,
                  (
                    m_inputParameters.m_useRasterCache 
                    &&
                    ( inputRasterPtr == m_inputParameters.m_feederRasterPtr->getCurrentObj() ) )
                  )
                )
                &&
                (
                  blenderInstance.blendIntoRaster1()
                )
              )
            )
          {
            const std::string message = "Error blending input raster " +
              std::to_string( inputRasterIdx );
            if( m_inputParameters.m_abortOnFirstError )
            {
              TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
            }
            else
            {
              TERP_INSTANCE_LOGWARN( message );
              ++centerDist2RasterIdxMapIt;
              continue;
            }
          }
        }

        // updating the  gloabal mosaic boxes
        
        if( mosaicValidAreasUnionPtr.get() == nullptr )
        {
          mosaicValidAreasUnionPtr.reset(
            currRasterValidDataDelimiterMultiPolPtr.release() );
        }
        else
        {
          std::unique_ptr< te::gm::Geometry > boxesUnionResultPtr; // under the mosaic SRID

          try
          {
            boxesUnionResultPtr.reset( mosaicValidAreasUnionPtr->Union(
              currRasterValidDataDelimiterMultiPolPtr.get() ) );
          }
          catch( const std::exception& e )
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Mosaic bounding boxes union error : "
              + std::string( e.what() ) );
          }
          catch( ... )
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Mosaic bounding boxes union error" );
          }          
          
          TERP_INSTANCE_TRUE_OR_RETURN_FALSE( boxesUnionResultPtr.get() !=
            nullptr, "Mosaic bounding boxes union error" );
          
          boxesUnionResultPtr->setSRID( outputRasterPtr->getSRID() );

          if( 
              ( boxesUnionResultPtr->getGeomTypeId() == te::gm::MultiPolygonType )
              ||
              ( boxesUnionResultPtr->getGeomTypeId() == te::gm::MultiPolygonZType )
              ||
              ( boxesUnionResultPtr->getGeomTypeId() == te::gm::MultiPolygonMType )
              ||
              ( boxesUnionResultPtr->getGeomTypeId() == te::gm::MultiPolygonZMType )
            )
          {
             mosaicValidAreasUnionPtr.reset( static_cast<te::gm::MultiPolygon*>(boxesUnionResultPtr.release()) );
          }
          else if( 
                   ( boxesUnionResultPtr->getGeomTypeId() == te::gm::PolygonType )
                   ||
                   ( boxesUnionResultPtr->getGeomTypeId() == te::gm::PolygonZType )
                   ||
                   ( boxesUnionResultPtr->getGeomTypeId() == te::gm::PolygonMType )
                   ||
                   ( boxesUnionResultPtr->getGeomTypeId() == te::gm::PolygonZMType )
                 )
          {
            // transforming it into a te::gm::MultiPolygon
            te::gm::MultiPolygon* auxMultiPol = new te::gm::MultiPolygon( 0,
              te::gm::MultiPolygonType, boxesUnionResultPtr->getSRID(), nullptr );
            auxMultiPol->add( boxesUnionResultPtr.release() );
            auxMultiPol->computeMBR( true );
            assert( auxMultiPol->isValid() );

            mosaicValidAreasUnionPtr.reset( auxMultiPol );
          }
          else
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Invalid union geometry type" );
          }
        }
        
        // moving to the next raster

        ++centerDist2RasterIdxMapIt;
      }
      
      // reseting the output cache
      
      if( outputCachedRasterInstancePtr.get() ) outputCachedRasterInstancePtr.reset();
      
      return true;
    }

    void GeoMosaic::reset() _NOEXCEPT_OP(false)
    {
      Algorithm::reset();
      
      m_inputParameters.reset();
      m_isInitialized = false;
    }

    bool GeoMosaic::initialize( const AlgorithmInputParameters& inputParams )
      _NOEXCEPT_OP(false)
    {
      reset();

      GeoMosaic::InputParameters const* inputParamsPtr = dynamic_cast<
        GeoMosaic::InputParameters const* >( &inputParams );
      TERP_INSTANCE_TRUE_OR_THROW( inputParamsPtr, "Invalid paramters pointer" );

      m_inputParameters = *inputParamsPtr;

      // Checking the feeder

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_feederRasterPtr,
        "Invalid m_feederRasterPtr" )

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_feederRasterPtr->getObjsCount() > 0,
        "Invalid number of rasters" )

      // checking m_inputRastersBands

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        (static_cast<unsigned int>(m_inputParameters.m_inputRastersBands.size())) ==
        m_inputParameters.m_feederRasterPtr->getObjsCount(),
        "Bands mismatch" );

      for( std::vector< std::vector< unsigned int > >::size_type
        inputRastersBandsIdx = 0 ;  inputRastersBandsIdx <
        m_inputParameters.m_inputRastersBands.size() ; ++inputRastersBandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inputRastersBands[
          inputRastersBandsIdx ].size() > 0, "Invalid bands number" );

        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inputRastersBands[
          inputRastersBandsIdx ].size() ==  m_inputParameters.m_inputRastersBands[
          0 ].size(), "Bands number mismatch" );
      }
      
      // Checking eq type
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_equalizationType
        != InputParameters::InvalidEqType , "Invalid equalization type" );
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_localEqMaxNeighborRasters
        >= 1 , "Invalid parameter: local Equalization Max Neighbor Rasters" );      
      
      // checking target means and std devs
      
      if( m_inputParameters.m_equalizationType == InputParameters::TargetEqtype )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
          (
            ( 
              m_inputParameters.m_mosaicTargetMeans.size() 
              == 
              m_inputParameters.m_inputRastersBands[ 0 ].size() 
            )
            &&
            ( 
              m_inputParameters.m_mosaicTargetStdDevs.size() 
              == 
              m_inputParameters.m_inputRastersBands[ 0 ].size() 
            )          
            &&
            ( 
              m_inputParameters.m_mosaicTargetMeans.size()
              ==
              m_inputParameters.m_mosaicTargetStdDevs.size()
            )                    
          )
          , "Invalid target means or target standart deviation values" );
      }
      
      // Blender
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        ( 
          ( 
            m_inputParameters.m_blendMethod 
            ==
            te::rp::blender::RuleMethod 
          ) 
          ? 
          ( 
            m_inputParameters.m_blenderRulePtr 
            !=
            nullptr 
          ) 
          : 
          true 
        ),
        "Missing blender rule" );
      
      // finalizing

      m_isInitialized = true;

      return true;
    }

    bool GeoMosaic::isInitialized() const
    {
      return m_isInitialized;
    }

  } // end namespace rp
}   // end namespace te

